import threading
import queue
import inspect
import time


def poolable(pool_id=''):

    def mark(func):
        setattr(func, '__poolID__', pool_id)
        return func

    return mark


class ObjectWorkerThread(threading.Thread):

    def __init__(self, poolable_class, function_dict, pool_id, result_queue):
        super().__init__()

        self.poolable_object = poolable_class()
        self.result_queue = result_queue
        self.function_dict = function_dict

        self.active = False
        self.work = None
        self.result = None
        self.id = pool_id

        self.running = False

    def assign_work(self, work_parameter):
        self.active = True
        self.work = work_parameter

    def run(self):
        self.running = True
        while self.running:
            if self.work is not None:
                try:
                    self.result = self.function_dict[self.id](self.poolable_object, self.work)
                    self.result_queue.put(self.result)
                finally:
                    self.work = None
                    self.active = False
            else:
                time.sleep(0.01)


class ObjectWorkerPool:

    def __init__(self, poolable_class, pool_id='', thread_count=10):
        self.poolable_class = poolable_class
        self.id = pool_id

        self.function_dict = {}
        for name, func in inspect.getmembers(poolable_class):
            if callable(func) and hasattr(func, '__poolID__'):
                self.function_dict[func.__poolID__] = func

        self.result_queue = queue.Queue(100)
        self.workers = []
        for i in range(thread_count):
            worker = ObjectWorkerThread(
                self.poolable_class,
                self.function_dict,
                self.id,
                self.result_queue
            )
            self.workers.append(worker)
            worker.start()
        self.workload = []

    def assign_workload(self, workload_list):
        self.workload = workload_list

    def fetch(self):
        while not(len(self.workload) == 0 and self.result_queue.empty() and len(idle_workers) == len(self.workers)):
            # Getting the idle workers and assigning them new work
            idle_workers = list(filter(lambda x: not x.active, self.workers))

            for worker in idle_workers:
                if len(self.workload) != 0:
                    worker.assign_work(self.workload.pop())
                    idle_workers.remove(worker)

            if not self.result_queue.empty():
                yield self.result_queue.get()

            time.sleep(0.01)

    def close(self):
        for worker in self.workers:
            worker.running = False


class DictQuery:
    """
    This is a class, that is supposed to help with the processing of multi layered dictionary structures.
    This is a base class for a state object, that processes one such dictionary at a time, the dictionary being set or
    changed by the 'set' method. The class provides the query dict method, with which a nested value of the dict can
    be acquired by using a standard path query like 'key1/key2/key3'.
    It will be attempted to access the nested structure with this path, but in case there is an exception due to a none
    existing path or something else an exception will be caught and the method 'query_exception' will be executed,
    the query will then return whatever this method returns.
    There is the possibility to specify a default value for every query in case it fails and this is also passed to
    the exception method, wo it could be returned from there, for example

    :ivar dict: The currently processed dict of the object. All the queries will be applied to this dict until it is
        changed
    """
    def __init__(self):
        self.dict = None

    def process(self):
        raise NotImplementedError()

    def set(self, query_dict):
        """
        Sets/changes the currently processed dict of the object

        :param query_dict: The new dict to be processed
        :return: void
        """
        self.dict = query_dict

    def set_query_dict(self, query_dict):
        self.dict = query_dict

    def query_exception(self, query, query_dict, default):
        """
        This method is being called, when a query to the dict fails.

        :param query: The query string for which the exception occurred
        :param query_dict: The whole dict with which the exception occurred
        :param default: The default value given to the query method
        :return: The return will be the return of the query method in case of failure
        """
        raise NotImplementedError()

    def query(self, dict_query, default=None):
        return self.query_dict(dict_query, default)

    def query_dict(self, dict_query, default):
        """
        Attempts to extract the value to the given query "path" from the nested dictionary structure, but in case it
        fails the method 'query_exception' will be executed.

        :param dict_query: The path-like query string defining the order of keys to apply to the nested dict
        :param default: A possible default value in case the query fails
        :return: The value form the dict
        """
        try:
            keys = dict_query.split('/')
            current_dict = self.dict

            for key in keys:
                # Trying to turn the key into an int in case to acces a nested list
                try:
                    current_dict = current_dict[int(key)]
                except (ValueError, TypeError):
                    current_dict = current_dict[key]

            return current_dict
        except (KeyError, ValueError, TypeError) as excpetion:
            return self.query_exception(dict_query, self.dict, default)


if __name__ == '__main__':
    class A:

        def __init__(self):
            pass

        @poolable('ass')
        def add(self, tup):
            return tup[0] + tup[1]

        @poolable('nope')
        def sub(self, tup):
            return tup[0] - tup[1]

    pool = ObjectWorkerPool(A, 'nope', thread_count=3)
    pool.assign_workload([(1,3), (4,5), (34, 60)])
    for i in pool.fetch():
        print(i)
    pool.close()


import configparser
import os

# Changed 11.04.2018
# Added deprecated, because the 'config_singleton' method was to be re written, but did not want to delete it.
# Added pathlib because for the rewrite of the config_singleton functionality Path objects needed
from deprecated import deprecated
import pathlib


# Changed 11.04.2018
# This function, which was inspired to simplify the creation of config singletons in the ScopusWp project is now
# deprecated, due to me learning how to properly implement singletons in Python using duck typing

@deprecated
def config_singleton(project_path, config_name='config.ini'):
    """
    This function returns a class, which represents a singleton for access to the projects config file. The project
    path and name of the config file are saved to the class by passing them as parameters to the function.

    USAGE:
    In a local project a singleton access class can be created by creating a new class, which inherits from the return
    of this function. And no further methods are needed. The child class can be accessed by calling the 'get_instance'
    method to retrieve the ConfigParser object:

    EXAMPLE:
    Class Config(config_singleton("my/path", "config.ini")): pass
    Config.get_instance()
    # <configparser.ConfigParser>

    :param project_path: The path to the folder where the config file is located
    :param config_name: The string file name of the config file to be subject of the class. DEFAULT: config.ini
    :return: Class:Config
    """
    class Config:
        """
        A singleton class for the access to the config object
        """
        _instance = None

        def __init__(self):
            pass

        @staticmethod
        def get_instance():
            if Config._instance is None:
                Config._instance = Config._create_instance()

            return Config._instance

        @staticmethod
        def _create_instance():
            # Getting the path to the config file
            path = Config._config_path()
            # Creating the config parser object and returning it
            config = configparser.ConfigParser()
            config.read(path)

            return config

        @staticmethod
        def _config_path():
            # Joining the config path
            path = os.path.join(project_path, 'config.ini')
            return path

    return Config


def ConfigBase(path, filename='config.ini'):
    """
    CHANGELOG

    Added 11.04.2018
    Rewrite of the 'config_singleton' function, also a function that returns the base class for a custom config singleton
    but with the class actually being a singleton implemented using python magic methods/duck typing

    CHANGED 15.04.2018
    When using this function to generate a base class for singleton configs, one wants to be able to add constants to
    the config class, which will be used as global variables through out the project. There was a problem with that
    however, where because the get attr method was overwritten for the sake of duck typing, a class variable from the
    singleton class would not be found.
    So to the init of the singleton instance I added a for loop, which will add all upper case attributes of the class
    also as attributes to the singelton instance thus enabling access to those fields.

    Changed 15.04.2018
    Added a folder_path object attribute, which is being used as a parameter of the path function for the derive method
    as that just makes more sense to pass the folder path and filename individually

    :param path:
    :param filename:
    :return:
    """

    class Config:

        class Config:

            def __init__(self):
                self.filename = filename
                self.folder_path = pathlib.Path(path)
                print(str(self.folder_path))
                self.path = self.folder_path / filename
                self.dict = configparser.ConfigParser()
                self.dict.read(str(self.path))

            def load(self):
                self.dict = configparser.ConfigParser()
                self.dict.read(str(self.path))

            def save(self):
                self.dict.write(self.path)

            def __getitem__(self, item):
                return self.dict[item]

            def __setitem__(self, key, value):
                self.dict[key] = value

        _instance = None
        _parent = None

        def __init__(self):
            if self._instance is None:
                config = Config.Config()
                setattr(Config, '_instance', config)
                # Changed 15.04.2018: Adding all the constants, that were added as class fields to the singleton
                # class to the actual singleton instance

                for item in dir(self):
                    if item.isupper():
                        setattr(self._instance, item, getattr(self, item))

        def __getitem__(self, item):
            return self._instance[item]

        def __setitem__(self, key, value):
            self._instance[key] = value

        def __setattr__(self, key, value):
            setattr(self._instance, key, value)

        def __getattr__(self, item):
            return getattr(self._instance, item)

        def derive(self, name, path_function):
            """

            :param name:
            :param path_function:
            :return:
            """
            assert callable(path_function)

            #print(path_function(name, str(self.folder_path), self.filename))
            path, filename = path_function(name, str(self.folder_path), self.filename)
            config = ConfigBase(path, filename)
            setattr(config, '_parent', self._instance)

            return config

        @property
        def parent(self):
            return self._parent

    return Config





from sqlalchemy import and_, or_


def filter_dict(model_class, parameter_dict):
    filter_parameter_dict = {}
    for key, value in parameter_dict.items():
        filter_parameter_dict['model_class.{}'.format(key)] = value
    return filter_parameter_dict


def filter_string_junction(model_class, parameter_dict, junction_string, expression):
    """
    CHANGELOG

    CHanged 19.04.2018
    Replacing a double quote with a single quote whenever dealing with a string value because when the string value
    e.g. a title contains a double quote naturally that lead to an error
    :param model_class:
    :param parameter_dict:
    :param junction_string:
    :param expression:
    :return:
    """
    filter_parameter_dict = filter_dict(model_class, parameter_dict)
    string_list = [junction_string, '(']
    for key, value in filter_parameter_dict.items():
        try:
            value_string = '{}'.format(int(value))
        except:
            value_string = '"{}"'.format(value.replace('"', "'"))
        expression_string = expression.format(key, value_string)
        string_list.append(expression_string)
        string_list.append(',')
    string_list[-1] = ')'
    filter_string = ''.join(string_list)
    return filter_string


def filter_string_and(model_class, parameter_dict, expression='{}=={}'):
    return filter_string_junction(
        model_class,
        parameter_dict,
        'and_',
        expression
    )


def filter_string_or(model_class, parameter_dict, expression='{}=={}'):
    return filter_string_junction(
        model_class,
        parameter_dict,
        'or_',
        expression
    )


def get_or_create(model_class, parameter_dict):
    # Attempting to get the model from the database
    filter_string = filter_string_and(model_class, parameter_dict)
    a = model_class.query.filter(eval(filter_string)).first()
    if a is not None:
        return a
    b = model_class(**parameter_dict)
    return b


def add_elements(list_attribute, model_list):
    for model in model_list:
        list_attribute.append(model)